"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.dataLayer = dataLayer;
exports.dataLayerClass = dataLayerClass;

var _Grid = require("./Grid.js");

var _Trajectory = require("./Trajectory.js");

var _Point = require("./Point.js");

var _PointSeries = require("./PointSeries.js");

var _PointCollection = require("./PointCollection.js");

var _VerticalProfile = require("./VerticalProfile.js");

var _VerticalProfileCollection = require("./VerticalProfileCollection.js");

var _MultiPolygon = require("./MultiPolygon.js");

var _PolygonSeries = require("./PolygonSeries.js");

var _MultiPolygonSeries = require("./MultiPolygonSeries.js");

var _constants = require("../util/constants.js");

var _DOMAIN_LAYER_CLASSES, _COLLECTION_LAYER_CLA;

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var DOMAIN_LAYER_CLASSES = (_DOMAIN_LAYER_CLASSES = {}, _defineProperty(_DOMAIN_LAYER_CLASSES, _constants.COVJSON_GRID, _Grid.Grid), _defineProperty(_DOMAIN_LAYER_CLASSES, _constants.COVJSON_POINT, _Point.Point), _defineProperty(_DOMAIN_LAYER_CLASSES, _constants.COVJSON_POINTSERIES, _PointSeries.PointSeries), _defineProperty(_DOMAIN_LAYER_CLASSES, _constants.COVJSON_VERTICALPROFILE, _VerticalProfile.VerticalProfile), _defineProperty(_DOMAIN_LAYER_CLASSES, _constants.COVJSON_TRAJECTORY, _Trajectory.Trajectory), _defineProperty(_DOMAIN_LAYER_CLASSES, _constants.COVJSON_MULTIPOLYGON, _MultiPolygon.MultiPolygon), _defineProperty(_DOMAIN_LAYER_CLASSES, _constants.COVJSON_POLYGONSERIES, _PolygonSeries.PolygonSeries), _defineProperty(_DOMAIN_LAYER_CLASSES, _constants.COVJSON_MULTIPOLYGONSERIES, _MultiPolygonSeries.MultiPolygonSeries), _DOMAIN_LAYER_CLASSES);
var COLLECTION_LAYER_CLASSES = (_COLLECTION_LAYER_CLA = {}, _defineProperty(_COLLECTION_LAYER_CLA, _constants.COVJSON_POINT, _PointCollection.PointCollection), _defineProperty(_COLLECTION_LAYER_CLA, _constants.COVJSON_VERTICALPROFILE, _VerticalProfileCollection.VerticalProfileCollection), _COLLECTION_LAYER_CLA);
/**
 * Creates a layer for a given coverage data object or throws an error if no layer class could be found.
 * 
 * This is a convenience function over using {@link dataLayerClass} and manually
 * instantiating the layer.
 * 
 * @example
 * var cov = ...
 * var layer = C.dataLayer(cov, {parameter: 'temperature'}).addTo(map)
 *  
 * @return {function} A function fn(cov, options) which returns a new layer for
 *   the given coverage data object and which is initialized with the given layer options.
 * @throws {Error} If no layer class could be found.
 */

function dataLayer(cov, options) {
  var clazz = dataLayerClass(cov);

  if (!clazz) {
    var coll = cov.type === _constants.COVERAGECOLLECTION ? 'collection ' : '';
    throw new Error("No ".concat(coll, "layer class found for domainType=").concat(cov.domainType));
  }

  return new clazz(cov, options);
}
/**
 * Return a layer class usable for the given coverage data object,
 * or `undefined` if none was found.
 * If multiple layers match, then an arbitrary one is returned.
 *  
 * @example
 * var cov = ...
 * var clazz = C.dataLayerClass(cov)
 * if (clazz) {
 *   var layer = new clazz(cov, {parameter: 'temperature'}).addTo(map)
 * }
 * 
 * @param {object} cov The coverage data object.
 * @return {class|undefined} The layer class.
 */


function dataLayerClass(cov) {
  if ((cov.type === _constants.COVERAGE || cov.type === _constants.DOMAIN) && cov.domainType in DOMAIN_LAYER_CLASSES) {
    return DOMAIN_LAYER_CLASSES[cov.domainType];
  }

  if (cov.type === _constants.COVERAGECOLLECTION && cov.domainType in COLLECTION_LAYER_CLASSES) {
    return COLLECTION_LAYER_CLASSES[cov.domainType];
  }
}